# -*- coding: utf-8 -*-

import __builtin__
import mox
import path
import yaml

from . import core
from . import client
from . import test_helper
from .test_helper import unittest


class TestWatsonClient(test_helper.TestBase):

    def test_watch(self):
        working_dir = path.path(__file__).dirname()
        working_dir /= '../fixtures/project1/some_dir'
        project_dir = unicode((working_dir / '..').abspath())

        config = core.load_config(working_dir / '..' / core.CONFIG_FILENAME)

        cl = client.WatsonClient()
        cl.add_project = self.mox.CreateMockAnything()
        cl.add_project(project_dir, config)

        self.mox.ReplayAll()

        # when...
        cl.watch(working_dir)

        self.mox.VerifyAll()

    def test_watch_raise_WatsonError_without_config(self):
        cl = client.WatsonClient()
        working_dir = (path.path(__file__).dirname()
                       / '../fixtures/project_no_config')

        with self.assertRaises(core.WatsonError):
            cl.watch(working_dir)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-

import its
import logging
import mox

try:
    import unittest2 as unittest
except ImportError:
    if not its.py27:
        raise
    import unittest


logging.basicConfig(level=logging.DEBUG, format='%(levelname)s %(message)s')


class TestBase(unittest.TestCase):

    def setUp(self):
        self.mox = mox.Mox()

    def tearDown(self):
        self.mox.UnsetStubs()
        self.mox.ResetAll()



# -*- coding: utf-8 -*-

import collections
import its
import mox
import path
import SimpleXMLRPCServer
import tempfile

from fabric import context_managers
from fabric import operations
from watchdog import observers

from . import core
from . import test_helper
from .test_helper import unittest


class TestFindProjectDirectory(unittest.TestCase):

    def test_simple(self):
        look_for = ['requirements.txt']
        start = path.path(__file__).dirname()

        directory = core.find_project_directory(start, look_for=look_for)

        self.assertEqual(start.parent, directory)

    def test_WatsonError_dir_is_not_part_of_project(self):
        with self.assertRaises(core.WatsonError):
            _ = core.find_project_directory(tempfile.gettempdir())


class HeadlessProjectWatcher(core.ProjectWatcher):

    def _create_notification(self):
        pass

    def _show_notification(self, status):
        self._last_status = status

    def _hide_notification(self):
        pass


class TestProjectWatcher(test_helper.TestBase):

    @classmethod
    def setUpClass(self):
        self._project_dir = core.find_project_directory(
            path.path(__file__).dirname())

    def setUp(self):
        super(TestProjectWatcher, self).setUp()

        self.directory = 'a directory'
        self.watch = 'a watch'

        self.observer_mock = self.mox.CreateMock(observers.Observer)
        self.worker_mock = self.mox.CreateMock(core.ProjectBuilder)
        self.scheduler_mock = self.mox.CreateMock(core.EventScheduler)

        (self.observer_mock.schedule(
            mox.IsA(core.ProjectWatcher), path=self.directory, recursive=True)
            .AndReturn(self.watch))

    def get_watcher(self, config=None):
        return HeadlessProjectWatcher(
            config or {}, self.directory, self.scheduler_mock,
            self.worker_mock, self.observer_mock)

    def test_init(self):
        self.mox.ReplayAll()

        self.get_watcher()

        self.mox.VerifyAll()

    def test_shutdown(self):
        self.observer_mock.unschedule(self.watch)
        self.mox.ReplayAll()

        self.get_watcher().shutdown()

        self.mox.VerifyAll()

    def test_build(self):
        status = (True, None)
        (self.worker_mock.execute_script(self.directory, ['nosetests'])
            .AndReturn(status))
        self.mox.ReplayAll()

        watcher = self.get_watcher({'name': 'test', 'script': ['nosetests']})
        watcher.build()

        self.mox.VerifyAll()
        self.assertEqual(status, watcher._last_status)


class HeadlessWatsonServer(core.WatsonServer):

    def _init_pynotify(self):
        pass


class TestWatsonServer(test_helper.TestBase):

    def setUp(self):
        super(TestWatsonServer, self).setUp()

        self.mox.StubOutClassWithMocks(SimpleXMLRPCServer,
                                       "SimpleXMLRPCServer")

        hostport = ("localhost", 0x221B)
        self.server_mock = SimpleXMLRPCServer.SimpleXMLRPCServer(
            hostport, allow_none=True)
        self.server_mock.register_instance(mox.IsA(core.WatsonServer))

        self.mox.StubOutClassWithMocks(observers, "Observer")
        self.observer_mock = observers.Observer()

        self.mox.StubOutClassWithMocks(core, "EventScheduler")
        self.scheduler_mock = core.EventScheduler()

    def test_init(self):
        self.mox.ReplayAll()

        HeadlessWatsonServer()

        self.mox.VerifyAll()

    def test_start(self):
        self.server_mock.serve_forever()
        self.observer_mock.start()
        self.scheduler_mock.start()

        self.mox.ReplayAll()

        HeadlessWatsonServer()._start()

        self.mox.VerifyAll()

    def test_shutdown(self):
        self.server_mock.server_close()
        self.observer_mock.stop()
        self.observer_mock.join()
        self.scheduler_mock.stop()
        self.scheduler_mock.join()

        self.mox.ReplayAll()

        HeadlessWatsonServer().shutdown()

        self.mox.VerifyAll()

    def test_hello(self):
        self.mox.ReplayAll()

        version = HeadlessWatsonServer().hello()

        self.mox.VerifyAll()
        self.assertIn(core.VERSION, version)


class ResultMock(collections.namedtuple('ResultMock', ['succeeded', 'msg'])):
    pass


class TestProjectBuilder(test_helper.TestBase):

    def setUp(self):
        super(TestProjectBuilder, self).setUp()

        self.working_dir = 'a directory'

        lcd_mock = self.mox.CreateMockAnything()
        lcd_mock.__enter__()
        lcd_mock.__exit__(mox.IgnoreArg(), mox.IgnoreArg(), mox.IgnoreArg())
        self.mox.StubOutWithMock(context_managers, 'lcd')
        context_managers.lcd(self.working_dir).AndReturn(lcd_mock)

    def _stubout_local(self, script, results):
        self.mox.StubOutWithMock(operations, 'local')
        for command, result in zip(script, results):
            operations.local(command, capture=True).AndReturn(
                ResultMock(result, command))

    def test_execute_script_internal(self):
        script = ['echo 1', 'echo 2']
        self._stubout_local(script, [True, True])

        self.mox.ReplayAll()

        worker = core.ProjectBuilder()
        result = worker._execute_script_internal(self.working_dir, script)

        self.mox.VerifyAll()
        self.assertEqual((True, ResultMock(True, script[-1])), result)

    def test_execute_runs_until_first_failure(self):
        script = ['echo 1', 'echo 2', 'echo 3']

        # only first two commands are passed to operations.local
        self._stubout_local(script[:2], [True, False])
        self.mox.ReplayAll()

        worker = core.ProjectBuilder()
        result = worker._execute_script_internal(self.working_dir, script)

        self.mox.VerifyAll()
        self.assertEqual((False, ResultMock(False, script[1])), result)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-

from __future__ import absolute_import

import os
import path
import sys
import time

from daemon import runner

from . import core


WATSON_DIR = path.path('~/.watson').expand()


class _DaemonRunner(runner.DaemonRunner):
    """Modified DaemonRunner that checks pidfile before opening a context."""

    def parse_args(self, args=None):
        self.action_funcs[u'start'] = self._start
        self.action_funcs[u'stop'] = self._stop
        self.action_funcs[u'restart'] = self._restart

    def _start(self):
        if self.pidfile.read_pid() is not None:
            raise runner.DaemonRunnerStartFailureError(
                u"PID file %r already locked" % self.app.pidfile_path)

        return super(_DaemonRunner, self)._start()

    def do_action(self, action=None):
        self.action = action
        func = self._get_action_func()
        func()


class WatsonDaemon(object):

    def __init__(self):
        self.pidfile_path = WATSON_DIR / 'pid'
        self.stdin_path = path.path('/dev/null')
        self.stdout_path = WATSON_DIR / 'stdout'
        self.stderr_path = WATSON_DIR / 'stderr'
        self.pidfile_timeout = 0

        # Create watson directory if it is not already there
        WATSON_DIR.mkdir_p()

    def run(self):
        server = None
        try:
            server = core.WatsonServer()
            server._start()
        except KeyboardInterrupt:
            pass
        finally:
            if server is not None:
                server.shutdown()
                server._join()

    def perform(self, action, fork=False):
        if not fork or os.fork() == 0:
            _DaemonRunner(self).do_action(action)
        elif fork:
            time.sleep(1)


def main():
    if len(sys.argv) < 2:
        return

    command = sys.argv[1]

    if command in ['run']:
        WatsonDaemon().run()

if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

import logging
import path
import socket
import sys
import xmlrpclib


from . import core
from . import daemon


class WatsonClient(xmlrpclib.ServerProxy):

    def __init__(self):
        self.endpoint = ('localhost', 0x221B)
        xmlrpclib.ServerProxy.__init__(self, 'http://%s:%s/' % self.endpoint,
                                       allow_none=True)

    def watch(self, working_dir="."):
        project_dir = core.find_project_directory(working_dir)
        config_file = project_dir / core.CONFIG_FILENAME
        config = core.load_config(config_file)

        # TODO(dejw): write a test for marshaling path.path objects
        self.add_project(unicode(project_dir), config)


def main():
    """usage: watson watch

    Repository watcher - watches for filesystem changes of your project and
    constantly builds it and keeps you posted about the build status.

    Commands:

      watch     starts watching a project or updates its status if it was
                already being watched

    """
    if len(sys.argv) < 2:
        print main.__doc__.strip()
        return

    command = sys.argv[1]

    if command in ['start', 'stop', 'restart']:
        daemon.WatsonDaemon().perform(command, fork=True)

    if command == 'watch':
        client = WatsonClient()

        try:
            version = client.hello()
        except socket.error:
            logging.warning('Could not connect to Watson server; Spawning one')
            daemon.WatsonDaemon().perform('start', fork=True)
            try:
                version = client.hello()
            except socket.error:
                raise core.WatsonError('Could not connect to the local watson '
                                       'server at %s' % (client.endpoint,))

        client.watch()

if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

from __future__ import absolute_import

import atexit
import logging
import os
import path
import re
import SimpleXMLRPCServer
import sched
import threading
import time
import yaml

from fabric import context_managers
from fabric import decorators
from fabric import operations
from multiprocessing import pool
from stuf import collects
from watchdog import events
from watchdog import observers

from . import VERSION


logging.basicConfig(level=logging.DEBUG, format='%(levelname)s %(message)s')


CONFIG_FILENAME = '.watson.yaml'
DEFAULT_PROJECT_INDICATORS = [CONFIG_FILENAME, '.vip', 'setup.py']

DEFAULT_GLOBAL_CONFIG_FILE = path.path('~/.watson/config.yaml').expand()
DEFAULT_CONFIG = {
    'endpoint': 'localhost:%s' % 0x221B,
    'ignore': ['.git/.*', '.*.pyc'],
    'build_timeout': 3
}


class WatsonError(StandardError):
    pass


def find_project_directory(start=".", look_for=None):
    """Finds a directory that looks like a project directory.

    The search is performed up in the directory tree, and is finished when
    one of the terminators is found.

    Args:
        start: a path (directory) from where the search is started
            "." by default
        look_for: a list of search terminators,
            core.DEFAULT_PROJECT_INDICATORS by default

    Returns:
        A path to a directory that contains one of terminators

    Raises:
        WatsonError: when no such directory can be found
    """
    look_for = set(look_for or DEFAULT_PROJECT_INDICATORS)

    directory = path.path(start).abspath()

    while directory.parent != directory:
        items = os.listdir(directory)
        if any(i in look_for for i in items):
            return directory

        directory = directory.parent

    raise WatsonError('%s does not look like a project subdirectory' % start)


def get_project_name(working_dir):
    """Returns a project name from given working directory."""
    return path.path(working_dir).name


def load_config(config_file):
    logging.info('Loading config: %s', config_file)
    config_file = path.path(config_file).abspath()
    project_dir = config_file.dirname()

    if not config_file.exists():
        raise WatsonError('config %s does not exist' % config_file)

    with open(config_file) as f:
        config = yaml.load(f)

    return config


def load_config_safe(config_file):
    try:
        return load_config(config_file)
    except WatsonError:
        return {}


class EventScheduler(threading.Thread):

    def __init__(self):
        threading.Thread.__init__(self)
        self._sched = sched.scheduler(time.time, self.delay)
        self._is_finished = False
        self._condition = threading.Condition()
        self._join_event = threading.Event()

    def delay(self, timeout):
        with self._condition:
            self._condition.wait(timeout)

    @property
    def is_finished(self):
        with self._condition:
            return self._is_finished

    def schedule(self, event, delay, function):
        with self._condition:
            logging.info('Scheduling %s in %ss', function.__name__, delay)
            if event is not None:
                self._sched.cancel(event)

            self._condition.notify()
            return self._sched.enter(delay, 1, function, [])

    def stop(self):
        with self._condition:
            logging.info('Stopping event scheduler')
            self._is_finished = True
            for event in self._sched.queue:
                self._sched.cancel(event)

            self._condition.notify()

    def join(self, timeout=None):
        self._join_event.wait(timeout)

    def run(self):
        logging.info('Starting event scheduler')

        while not self.is_finished:
            self._sched.run()
            with self._condition:
                if not self._is_finished:
                    logging.info('Queue is empty')
                    self._condition.wait()

        self._join_event.set()
        logging.info('Event scheduler stopped')


class Config(collects.ChainMap):

    _KEYS_TO_WRAP = ['ignore', 'script']

    def __init__(self, *configs):
        if not configs:
            configs.append(DEFAULT_CONFIG)
        super(Config, self).__init__(*configs)

    def __getitem__(self, item):
        value = super(Config, self).__getitem__(item)

        if item in self._KEYS_TO_WRAP and not isinstance(value, list):
            value = [value]

        return value

    def push(self, config):
        new_config = self.new_child()
        new_config.update(config)
        return new_config

    def replace(self, config):
        self.maps[0] = config

    def __getattr__(self, attr):
        return self.__getitem__(attr)


class ProjectWatcher(events.FileSystemEventHandler):

    # TODO(dejw): should expose some stats (like how many times it was
    #             notified) or how many times it succeeed in testing etc.

    def __init__(self, config, working_dir, scheduler, builder, observer):
        super(ProjectWatcher, self).__init__()

        self._event = None

        self.name = get_project_name(working_dir)
        self.working_dir = path.path(working_dir)
        self.set_config(config)

        self._last_status = (None, None)
        self._create_notification()

        self._scheduler = scheduler
        self._builder = builder
        self._observer = observer

        # TODO(dejw): allow to change observing patterns (and recursiveness)
        self._watch = observer.schedule(self, path=working_dir, recursive=True)

        logging.info('Observing %s', working_dir)

    def __repr__(self):
        return '<ProjectWatcher %s(%s)>' % (self.name, self.working_dir)

    @property
    def script(self):
        return self._config['script']

    def set_config(self, config):
        self._config = config

    def shutdown(self):
        logging.info('Shuting down project: %r', self)
        self._hide_notification()
        self._observer.unschedule(self._watch)

    def on_any_event(self, event):
        logging.debug('Event: %r', event)

        event_path = event.src_path[len(self.working_dir):].lstrip('/')
        for ignore in self._config['ignore']:
            if re.match(ignore, event_path):
                return

        # Automatically pickup config changes
        logging.debug(event_path)
        if event_path == CONFIG_FILENAME:
            self._config.replace(load_config(event.src_path))

        self.schedule_build()

    def schedule_build(self, timeout=None):
        """Schedules a building process in configured timeout."""

        if timeout is None:
            timeout = self._config['build_timeout']

        logging.debug('Scheduling a build in %ss', timeout)
        self._event = self._scheduler.schedule(
            self._event, timeout, self.build)

    def build(self):
        """Builds the project and shows notification on result."""
        logging.info('Building %s (%s)', self.name, self.working_dir)
        self._event = None
        status = self._builder.execute_script(self.working_dir, self.script)
        self._show_notification(status)

    def _create_notification(self):
        try:
            import pynotify
            self._notification = pynotify.Notification('')
            self._notification.set_timeout(3)
        except ImportError:
            pass

    def _hide_notification(self):
        self._notification.close()

    def _show_notification(self, status):
        succeeed, result = status
        output = '\n'.join([result.stdout.strip(), result.stderr.strip()])
        output = output

        if not succeeed:
            self._notification.update(
                'Build of %s has failed' % self.name, output, 'dialog-error')
        else:
            self._notification.update(
                'Build of %s was successful' % self.name, output,
                'dialog-apply')

        self._notification.show()
        self._last_status = status


class ProjectBuilder(object):

    def execute_script(self, working_dir, script):
        return self._execute_script_internal(working_dir, script)

    @decorators.with_settings(warn_only=True)
    def _execute_script_internal(self, working_dir, script):
        succeeded = True
        result = None

        with context_managers.lcd(working_dir):
            for command in script:
                result = operations.local(command, capture=True)
                succeeded = succeeded and result.succeeded
                if not succeeded:
                    break

        return (succeeded, result)


class WatsonServer(object):

    def __init__(self):
        self._config = Config(load_config_safe(DEFAULT_GLOBAL_CONFIG_FILE))
        self._projects = {}

        self._builder = ProjectBuilder()
        self._observer = observers.Observer()
        self._scheduler = EventScheduler()
        self._init_pynotify()

        # TODO(dejw): read (host, port) from config in user's directory
        self.endpoint = ('localhost', 0x221B)
        self._api = SimpleXMLRPCServer.SimpleXMLRPCServer(
            self.endpoint, allow_none=True)
        self._api.register_instance(self)

    def _start(self):
        logging.info('Server listening on %s' % (self.endpoint,))
        self._scheduler.start()
        self._observer.start()
        self._api.serve_forever()

    def _join(self):
        self._api.shutdown()

    def _init_pynotify(self):
        logging.info('Configuring pynotify')
        try:
            import pynotify
            pynotify.init('Watson')
            assert pynotify.get_server_caps() is not None
        except ImportError:
            logging.error('pynotify not found; notifications disabled')

    def hello(self):
        return 'Watson server %s' % VERSION

    def shutdown(self):
        logging.info('Shuting down')

        for project in self._projects.itervalues():
            project.shutdown()

        self._api.server_close()
        self._observer.stop()
        self._scheduler.stop()

        self._observer.join()
        self._scheduler.join()

    def add_project(self, working_dir, config):
        try:
            logging.info('Adding a project: %s', working_dir)

            project_name = get_project_name(working_dir)
            config = self._config.push(config)

            if project_name not in self._projects:
                self._projects[project_name] = ProjectWatcher(
                    config, working_dir, self._scheduler, self._builder,
                    self._observer)

            else:
                self._projects[project_name].set_config(config)

            self._projects[project_name].schedule_build(0)
        except:
            import traceback
            traceback.print_exc()



# -*- coding: utf-8 -*-

import sys
import os

from setuptools import setup, find_packages
from setuptools.command.install import install as _install


#  Build requirements list
requirements = open('requirements.txt').readlines()

POST_INSTALL_MESSAGE = """
Remember to install pynotify to be able to see system notifications about
build statuses!
"""

def get_info():
    import watson
    return watson.VERSION, open('README.markdown').read()

version, long_description = get_info()

class install(_install):
    def run(self):
        _install.run(self)
        print POST_INSTALL_MESSAGE.strip()


kw = dict(
    name='watson-ci',
    version=version,
    url='https://github.com/dejw/watson-ci/',
    license='BSD',
    author='Dawid Fatyga',
    author_email='dawid.fatyga@gmail.com',
    description='watson is a simple CI server that helps you build your projects continuously',
    long_description=long_description,
    packages=find_packages(),
    zip_safe=False,
    platforms='any',
    install_requires=requirements,
    system_requires=['pynotify'],
    entry_points={
        'console_scripts': [
            'watson = watson.client:main',
        ]
    },
    classifiers=[
        # http://pypi.python.org/pypi?%3Aaction=list_classifiers
        #'Development Status :: 1 - Planning',
        'Development Status :: 2 - Pre-Alpha',
        #'Development Status :: 3 - Alpha',
        #'Development Status :: 4 - Beta',
        #'Development Status :: 5 - Production/Stable',
        #'Development Status :: 6 - Mature',
        #'Development Status :: 7 - Inactive',
        'Intended Audience :: Developers',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Information Technology',
        'Intended Audience :: Science/Research',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: BSD License',
        'Operating System :: POSIX',
        'Operating System :: MacOS',
        'Operating System :: Unix',
        'Programming Language :: Python',
        'Topic :: System :: Systems Administration',
        'Topic :: Utilities'
    ]
)


if __name__ == '__main__':
    setup(cmdclass={'install': install}, **kw)
